// AUTHOR: Michael Gill
// School or Mathematics
// Monash University
// Clayton, Victoria, AU 3800
// email:  Michael.Gill@Monash.edu
//
// A set of program wide definitions and typename declarations. All header files
// should include this file to be ensure project consistency.
// Copyright (C) 2019  Michael Gill
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef LATIN10_SOLVER_SOLVER_HPP_
#define LATIN10_SOLVER_SOLVER_HPP_

#include <algorithm>
#include <iostream>
#include <string>
#include <vector>

// #include "diagonals/diagonals.hpp"
#include "io/file.hpp"

// Compute if two 128 bit integers agree in any position
bool conflict(__uint128_t a, __uint128_t b) { return (a & b) > 0; }

// Compute if two 128 bit integers agree in exactly one position.
bool intersect(__uint128_t a, __uint128_t b) {
  __uint128_t x = a & b;
  if (x == 0) {
    return false;
  }
  return (x & (uint128_t)(x - (uint128_t)1)) == 0;
}

// Switches on the early exit mode for cell tracking.
#define GOOD_ENOUGH_DETECTION 6

#ifdef GOOD_ENOUGH_DETECTION
#define GOOD_LIMIT 6
#endif

// #define DEBUG
// The arm estimator must be in debug mode.
#define ARM_ESTIMATOR

std::array<uint128_t, N * N> masks;

// Struct keeps track of the remaining viable cells in a list format.
struct CellTracker_List {
  std::vector<std::array<uint128_t, N * N>> sq_cells;
  std::vector<size_t> windows;

  CellTracker_List(size_t squares) : sq_cells(squares), windows(squares, N * N) {
    for (size_t j = 0; j < sq_cells.size(); j++) {
      for (size_t i = 0; i < N * N; i++) {
        sq_cells[j] = masks;
      }
    }
  }

  void update(uint128_t transversal, size_t square) {
    size_t pos = 0;
    size_t check = windows[square] - 10;

    while (pos < windows[square]) {
      if (conflict(transversal, sq_cells[square][pos])) {
        // uint128_t tmp = sq_cells[square][pos];
        // sq_cells[square][pos] = sq_cells[square][windows[square] - 1];
        // sq_cells[square][windows[square] - 1] = tmp;
        std::swap(sq_cells[square][pos], sq_cells[square][windows[square] - 1]);
        windows[square]--;
        if (windows[square] == check) {
          return;
        }
      } else {
        pos++;
      }
    }
  }

  void revert(uint128_t transversal, size_t square) { windows[square] += N; }

  std::vector<uint128_t> get_cells(size_t square) const {
    std::vector<uint128_t> cell_arr;
    cell_arr.reserve(windows[square]);

    for (size_t idx = 0; idx < windows[square]; idx++) {
      cell_arr.push_back(sq_cells[square][idx]);
    }

    return cell_arr;
  }
};

// Bit based cell tracker offers dramatic speedup over the generalised list
// case.
struct CellTracker_Bit {
  std::vector<uint128_t> bitList;
  const uint128_t active = ~(((((uint128_t)1) << 28) - 1) << 100);

  CellTracker_Bit(size_t squares) : bitList(squares, 0) {}

  void update(uint128_t transversal, size_t square) {
    bitList[square] = bitList[square] | transversal;
  }

  void revert(uint128_t transversal, size_t square) {
    bitList[square] = bitList[square] ^ transversal;
  }

  std::vector<uint128_t> get_cells(size_t square) const {
    std::vector<uint128_t> cell_arr;
    cell_arr.reserve(N*N);

    uint128_t cells = ~bitList[square] & active;

    uint128_t x = cells & -cells;
    while (x > 0) {
      cell_arr.push_back(x);
      cells = cells ^ x;
      x = cells & -cells;
    }

    return cell_arr;
  }
};

#ifdef LISTTRACKING
typedef CellTracker_List CellTracker;
#else
typedef CellTracker_Bit CellTracker;
#endif

// Solver options struct contains the definitions of multiple options that
// are available to the solver.
struct SolverOptions {
  CellTracker c_tracker;
  size_t sid = 0;

  SolverOptions(size_t squares) : c_tracker(squares) {}

  bool is_complete(const std::vector<std::vector<uint128_t>> &partial) {
    for (size_t square = 0; square < partial.size() - 1; square++) {
      if (partial[square].size() < N) {
        return false;
      }
    }

    return true;
  }

  // Determines which cell is in the current square has the smallest number
  // of diagonals incident
  std::vector<uint128_t> cellector(const std::vector<uint128_t> &diagonals,
                                   size_t square) {
    std::vector<uint128_t> active_cells = c_tracker.get_cells(square);
    std::vector<uint128_t> min_diags = diagonals;

#ifdef GOOD_ENOUGH_DETECTION
    size_t good_count = 0;
    size_t is_good = false;
#endif

    for (size_t a_idx = 0; a_idx < active_cells.size(); a_idx++) {
      std::vector<uint128_t> diags;
      diags.reserve(min_diags.size() + 1);

      for (size_t d_idx = 0; d_idx < diagonals.size(); d_idx++) {
        if (conflict(active_cells[a_idx], diagonals[d_idx])) {
          diags.push_back(diagonals[d_idx]);
        }

        if (diags.size() >= min_diags.size()) {
          break;
        }
      }

      if (diags.size() == 0) {
        return diags;
      }

      if (diags.size() < min_diags.size()) {
        min_diags = std::move(diags);
#ifdef GOOD_ENOUGH_DETECTION
        if (min_diags.size() < GOOD_ENOUGH_DETECTION) {
          is_good = true;
        }
      }

      if (is_good) {
        good_count++;
        if (good_count == GOOD_LIMIT) {
          return min_diags;
        }
#endif
      }
    }
    return min_diags;
  }

  // Define square selector (squarelector) for if constexpr structs. This is
  // allows us to quickly interchange different search strategies.

  // Partially complete square 3 before alternating square 1 and 2
  struct R2P3A12 {};
  // Alternate between all squares equally
  struct Afull {};
  // Alternate only between square 1 and 2
  struct A12 {};
  // Partially complete 3 then complete 1 then 2
  struct R2P3F1F2 {};

  template <typename Tag = Afull>
  size_t squarelector(const std::vector<std::vector<uint128_t>> &diagonals,
                      const std::vector<std::vector<uint128_t>> &partial) {

    if constexpr (std::is_same<Tag, R2P3A12>::value) {

      if (partial[2].size() < 2) {
        return 2;
      }

      if (partial[1].size() < partial[0].size()) {
        return 1;
      } else {
        return 0;
      }

    } else if constexpr (std::is_same<Tag, Afull>::value) {

      std::vector<size_t> indices(partial.size(), 0);
      std::iota(indices.begin(), indices.end(), 0);

      return *std::min_element(
          indices.begin(), indices.end(),
          [&partial](const auto &a, const auto &b) -> bool {
            return partial[a].size() < partial[b].size();
          });

    } else if constexpr (std::is_same<Tag, A12>::value) {

      if (partial[0].size() > partial[1].size()) {
        return 1;
      } else {
        return 0;
      }

    } else if constexpr (std::is_same<Tag, R2P3F1F2>::value) {

      if (partial[2].size() < 2) {
        return 2;
      }

      if (partial[0].size() < N) {
        return 0;
      }
      return 1;
    }
  }

  size_t next_id() { return sid++; }
};

// Construct a square from the diagonal
std::vector<Square> squareify(std::vector<std::vector<uint128_t>> &partial) {
  std::vector<Square> squares(partial.size());
  for (size_t square = 0; square < partial.size(); square++) {
    for (size_t cell = 0; cell < N * N; cell++) {
      for (size_t diag = 0; diag < partial[square].size(); diag++) {
        if (conflict(masks[cell], partial[square][diag])) {
          squares[square][cell] = diag + 1;
          break;
        }
      }
    }
  }
  return squares;
}


template <bool WriteSquares, bool SquareFirst>
size_t
solve(FileWriter &fw, const std::vector<std::vector<uint128_t>> &diagonals,
      std::vector<std::vector<uint128_t>> &partial, SolverOptions &s_options) {

  // If the partial is complete then we squareify it for printing.
  if (s_options.is_complete(partial)) {
    if constexpr (WriteSquares) {
      std::vector<Square> squares = squareify(partial);

      fw.write(squares, std::to_string(s_options.next_id()));
    }

    return 1;
  }

  size_t square;
  if constexpr (SquareFirst) {
    square = s_options.squarelector<SolverOptions::R2P3F1F2>(diagonals, partial);
  } else {
    square = s_options.squarelector<SolverOptions::R2P3A12>(diagonals, partial);
  }

  std::vector<uint128_t> cell_diags =
      s_options.cellector(diagonals[square], square);

#ifdef DEBUG
  if (partial[0].size() == 1) {
    std::cout << square << " " << cell_diags.size() << "\n";
  }
#endif
  size_t accumulator = 0;
  for (size_t d_idx = 0; d_idx < cell_diags.size(); d_idx++) {
    // Add the chosen diag to the partial.
    partial[square].push_back(cell_diags[d_idx]);
    s_options.c_tracker.update(cell_diags[d_idx], square);

    // We build a new set of diagonals.
    std::vector<std::vector<uint128_t>> new_diags(partial.size());
    for (size_t sq = 0; sq < partial.size(); sq++) {
      new_diags.reserve(diagonals[sq].size());

      // If we are looking at the relevant square then we want to eleminate any
      // conflicts. Conversly if we are looking at an off cycle square then we
      // want to keep everyhting that "intersects" but only exactly once.
      if (square == sq) {
        for (auto &diag : diagonals[square]) {
          if (!conflict(cell_diags[d_idx], diag)) {
            new_diags[square].push_back(diag);
          }
        }
      } else {
        for (auto &diag : diagonals[sq]) {
          if (intersect(cell_diags[d_idx], diag)) {
            new_diags[sq].push_back(diag);
          }
        }
      }
    }

    if (square < 2) {
      new_diags[2].clear();
    }

    accumulator +=
        solve<WriteSquares, SquareFirst>(fw, new_diags, partial, s_options);

    partial[square].pop_back();
    s_options.c_tracker.revert(cell_diags[d_idx], square);
#ifdef DEBUG
    if (partial[0].size() == 1) {
      std::cout << "Finished part " << d_idx + 1 << " of " << cell_diags.size()
                << " in " << square << "\n";
#ifdef ARM_ESTIMATOR
      if (square == 0) {
        exit(0);
      }
#endif
    }
#endif
  }

  return accumulator;
}

// This overload specifies a starting point using the modular trick.
template <bool WriteSquares, bool SquareFirst>
size_t solve(FileWriter &fw,
             const std::vector<std::vector<uint128_t>> &diagonals,
             std::vector<std::vector<uint128_t>> &partial,
             SolverOptions &s_options, size_t m_class, size_t mod) {

  // If the partial is complete then we squareify it for printing.
  if (s_options.is_complete(partial)) {
    if constexpr (WriteSquares) {
      std::vector<Square> squares = squareify(partial);

      fw.write(squares, std::to_string(s_options.next_id()));
    }

    return 1;
  }

  size_t square;
  if constexpr (SquareFirst) {
    square = s_options.squarelector<SolverOptions::R2P3F1F2>(diagonals, partial);
  } else {
    square = s_options.squarelector<SolverOptions::R2P3A12>(diagonals, partial);
  }

  std::vector<uint128_t> cell_diags =
      s_options.cellector(diagonals[square], square);

#ifdef DEBUG
  if (partial[0].size() == 1) {
    std::cout << square << " " << cell_diags.size() << "\n";
  }
#endif
  size_t accumulator = 0;
  for (size_t d_idx = 0; d_idx < cell_diags.size(); d_idx++) {
    if (d_idx % mod == m_class) {
      // Add the chosen diag to the partial.
      partial[square].push_back(cell_diags[d_idx]);
      s_options.c_tracker.update(cell_diags[d_idx], square);

      // We build a new set of diagonals.
      std::vector<std::vector<uint128_t>> new_diags(partial.size());
      for (size_t sq = 0; sq < partial.size(); sq++) {
        new_diags.reserve(diagonals[sq].size());

        // If we are looking at the relevant square then we want to eleminate
        // any conflicts. Conversly if we are looking at an off cycle square
        // then we want to keep everyhting that "intersects" but only exactly
        // once.
        if (square == sq) {
          for (auto &diag : diagonals[square]) {
            if (!conflict(cell_diags[d_idx], diag)) {
              new_diags[square].push_back(diag);
            }
          }
        } else {
          for (auto &diag : diagonals[sq]) {
            if (intersect(cell_diags[d_idx], diag)) {
              new_diags[sq].push_back(diag);
            }
          }
        }
      }

      accumulator +=
          solve<WriteSquares, SquareFirst>(fw, new_diags, partial, s_options);

      partial[square].pop_back();
      s_options.c_tracker.revert(cell_diags[d_idx], square);
#ifdef DEBUG
      if (partial[0].size() == 1) {
        std::cout << "Finished part " << d_idx + 1 << " of "
                  << cell_diags.size() << " in " << square << "\n";
      }
#endif
    }
  }

  return accumulator;
}

// Given a set of transversals, determine the total number of ways the
// transversals can be assembled into a set of mutually orthogonal latin
// squares subject to a set of conditions contained in a SolverOptions
template <bool HotStart = false, bool SquareFirst = false>
size_t solve(FileWriter &fw,
             const std::vector<std::vector<uint128_t>> &diagonals,
             const bool write_files, size_t m_class = 0, size_t mod = 1,
             uint128_t starter = 0) {
  for (uint128_t cell = 0; cell < N * N; cell++) {
    masks[cell] = (uint128_t)(1) << cell;
  }

  SolverOptions s_options(diagonals.size());

  std::vector<std::vector<uint128_t>> partial(diagonals.size());
  for (size_t square = 0; square < partial.size(); square++) {
    partial[square].reserve(N);
  }

  std::vector<std::vector<uint128_t>> new_diags(partial.size());
  if constexpr (HotStart) {
    if (starter == 0) {
      starter = diagonals[0][0];
    }

    partial[0].push_back(starter);

    s_options.c_tracker.update(starter, 0);

    // We build a new set of diagonals.
    for (size_t sq = 0; sq < partial.size(); sq++) {
      new_diags.reserve(diagonals[sq].size());

      // If we are looking at the relevant square then we want to eliminate
      // any conflicts. Conversely if we are looking at an off cycle square
      // then we want to keep everything that "intersects" but only exactly
      // once.
      if (0 == sq) {
        for (auto &diag : diagonals[0]) {
          if (!conflict(starter, diag)) {
            new_diags[0].push_back(diag);
          }
        }
      } else {
        for (auto &diag : diagonals[sq]) {
          if (intersect(starter, diag)) {
            new_diags[sq].push_back(diag);
          }
        }
      }
    }
  } else {
    new_diags = diagonals;
  }

  if (write_files) {
    return solve<true, SquareFirst>(fw, new_diags, partial, s_options, m_class,
                                    mod);
  } else {
    return solve<false, SquareFirst>(fw, new_diags, partial, s_options, m_class,
                                     mod);
  }
}

#endif // LATIN10_SOLVER_SOLVER_HPP_
