// AUTHOR: Michael Gill
// School or Mathematics
// Monash University
// Clayton, Victoria, AU 3800
// email:  Michael.Gill@Monash.edu
//
// A set of program wide definitions and typename declarations. All header files
// should include this file to be ensure project consistency.
// Copyright (C) 2019  Michael Gill
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "io/file.hpp"

FileReader::FileReader(std::istream& stream, FileOptions& opt) :
  stream(stream), options(opt)
{
  switch (options.enc) {
    case encoding::NUMERIC :
      d = new NumericDecoder;
      break;
    case encoding::ALPHA :
      d = new AlphaDecoder;
      break;
    case encoding::BITPACKED :
      d = new BitDecoder;
      break;
    case encoding::FACTORIAL :
      d = new FactorialDecoder;
      break;
  }

  std::string line;
  if (!std::getline(stream, line)) {
    options.valid = false;
    return;
  }

  for (size_t i = 0; i < line.size(); i++) {
    options.gSizes.push_back((size_t)(line[i]-'0'));
  }

  for (size_t i = 0; i < GDD_B_SIZES; i++) {
    stream  >> options.bSizes[i];
  }

  stream >> options.rIdx >> options.cIdx;

  // options.s1 = 0;
  // options.s2 = 1;
  // if (options.gSizes.size() > 4) {
  //   stream >> options.s1 >> options.s2;
  //   std::getline(stream, line);
  // }

  std::getline(stream, line);
  std::getline(stream, line);
  if (line != "BEGIN") {
    options.valid = false;
  }

  track(options.offset);

  options.valid = true;
}

FileReader::~FileReader() {
  delete d;
}

std::string FileReader::nextID() const {
  return sqID;
}

bool FileReader::hasNext() const {
  return sqID != "END" && options.valid;
}

void FileReader::read(std::vector<Square>& squares) {
  for (size_t i = 0; i < options.squares; i++) {
    d->decode(stream, squares[i]);
  }

  trackNext();
}

std::string FileReader::readAsString() {
  std::stringstream ss;
  std::string line;
  for (size_t i = 0; i < options.squares-1; i++) {
    std::getline(stream, line);
    ss << line << "\n";
  }

  std::getline(stream, line);
  ss << line;

  trackNext();

  return ss.str();
}

void FileReader::trackNext() {
  for (size_t i = 0; (i + 1 < options.skip) && hasNext(); i++) {
    std::getline(stream, sqID);
    std::string junk;
    for (size_t j = 0; j < options.squares; j++) {
      std::getline(stream, junk);
    }
  }

  std::getline(stream, sqID);
}

void FileReader::track(size_t t) {
  for (size_t i = 0; (i < t) && hasNext(); i++) {
    std::getline(stream, sqID);
    std::string junk;
    for (size_t j = 0; j < options.squares; j++) {
      std::getline(stream, junk);
    }
  }

  std::getline(stream, sqID);
}

FileWriter::FileWriter(std::ostream& stream, FileOptions& opt) :
  stream(stream),
  options(opt),
  idPrefix("")
{
  switch (options.enc) {
    case encoding::NUMERIC :
      e = new NumericEncoder;
      break;
    case encoding::ALPHA :
      e = new AlphaEncoder;
      break;
    case encoding::BITPACKED :
      e = new BitEncoder;
      break;
    case encoding::FACTORIAL :
      e = new FactorialEncoder;
      break;
  }

  stream << join(options.gSizes.begin(), options.gSizes.end(), "");
  stream << "\n";

  stream << join(options.bSizes.begin(), options.bSizes.end());
  stream << "\n";

  stream << options.rIdx << " " << options.cIdx << "\n";

  if (options.gSizes.size() > 4) {
    stream << options.ms1 << " " << options.ms2 << "\n";
  }

  stream << "BEGIN" << std::endl;
}

FileWriter::FileWriter(
  std::ostream& stream,
  FileOptions& opt,
  std::string idPrefix) :
  stream(stream),
  options(opt),
  idPrefix(idPrefix)
{
  switch (options.enc) {
    case encoding::NUMERIC :
      e = new NumericEncoder;
      break;
    case encoding::ALPHA :
      e = new AlphaEncoder;
      break;
    case encoding::BITPACKED :
      e = new BitEncoder;
      break;
    case encoding::FACTORIAL :
      e = new FactorialEncoder;
      break;
  }

  stream << join(options.gSizes.begin(), options.gSizes.end(), "");
  stream << "\n";

  stream << join(options.bSizes.begin(), options.bSizes.end());
  stream << "\n";

  stream << options.rIdx << " " << options.cIdx << "\n";

  // if (options.gSizes.size() > 4) {
  //   stream << options.ms1 << " " << options.ms2 << "\n";
  // }

  stream << "BEGIN" << std::endl;
}

FileWriter::~FileWriter() {
  delete e;
  stream << "END" << std::endl;
}

void FileWriter::write(const std::vector<Square>& squares, const std::string& id) {
  if (idPrefix != "") {
    stream << idPrefix << "-";
  }
  stream << id << "\n";

  for (size_t i = 0; i < options.squares; i++) {
    e->encode(stream, squares[i]);
    stream << "\n";
  }
}

void FileWriter::writeAsString(const std::string& squares, const std::string& id) {
  if (idPrefix != "") {
    stream << idPrefix << "-";
  }
  stream << id << "\n";
  stream << squares << "\n";
}

TemplateReader::TemplateReader(std::istream& input, TemplateOptions& opts) :
  stream(input),
  options(opts)
{
  std::string line;
  if (!std::getline(stream, line)) {
    options.valid = false;
    return;
  }

  for (size_t i = 0; i < line.size(); i++) {
    options.gSizes.push_back((size_t)(line[i]-'0'));
  }

  options.squares = options.gSizes.size() - 2;

  for (size_t i = 0; i < GDD_B_SIZES; i++) {
    stream  >> options.bSizes[i];
  }

  stream >> options.rIdx >> options.cIdx;
  std::getline(stream, line);

  std::getline(stream, line);
  if (line != "BEGIN") {
    options.valid = false;
  }

  std::getline(stream, template_id);
}

TemplateReader::~TemplateReader(){}


bool TemplateReader::hasNext() const {
  return template_id != "END" && options.valid;
}

std::string TemplateReader::nextID() const {
  return template_id;
}

void TemplateReader::skip(size_t i) {
  std::string garbage;
  for (size_t count = 0; count < i && hasNext(); count++) {
    for (size_t row = 0; row < N; row++) {
      std::getline(stream, garbage);
    }

    std::getline(stream, template_id);
  }
}

TemplateWriter::TemplateWriter(std::ostream& stream, TemplateOptions& opt) :
  TemplateWriter(stream, opt, "")
{}

  // template <bool HotStart>
  // void TemplateReader::read(Template& temp);
  // Implemented in header.

TemplateWriter::TemplateWriter(
  std::ostream& stream,
  TemplateOptions& opt,
  std::string idPrefix) :
  stream(stream),
  options(opt),
  idPrefix(idPrefix)
{
  stream << join(options.gSizes.begin(), options.gSizes.end(), "");
  stream << "\n";

  stream << join(options.bSizes.begin(), options.bSizes.end());
  stream << "\n";

  stream << options.rIdx << " " << options.cIdx << "\n";

  stream << "BEGIN" << std::endl;
}

TemplateWriter::~TemplateWriter() {
  stream << "END" << std::endl;
}

void TemplateWriter::write(const std::vector<Square>& squares, const std::string& id) {
  if (idPrefix != "") {
    stream << idPrefix << "-";
  }
  stream << id << "\n";

  for (size_t row = 0; row < N; row++) {
    for (size_t col = 0; col < N; col++) {
      size_t val = 0;
      for (size_t group = 0; group < squares.size(); group++) {
        if (squares[group][row*N + col]) {
          val = val | (1 << group);
        }
      }
      stream << itohexc(val);
    }
    stream <<"\n";
  }
}
