// AUTHOR: Michael Gill
// School or Mathematics
// Monash University
// Clayton, Victoria, AU 3800
// email:  Michael.Gill@Monash.edu
//
// Functions for dealing with answers relating to transversals.
// Example: [Find the number of common transversals,
//           Find the largest set of common disjoint transversals]
// Copyright (C) 2019  Michael Gill
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <algorithm>
#include <fstream>
#include <iostream>
#include <string>

#include "diagonals/diagonals.hpp"
#include "io/cloptions.hpp"
#include "io/file.hpp"
#include "solver/solver.hpp"

using namespace std;

// Powers of 2, computed at compile time through const expr
constexpr size_t pow2(size_t n) {
  if (n == 0) {
    return 1;
  } else {
    const size_t val = pow2(n / 2);
    if (n % 2) {
      return 2 * val * val;
    } else {
      return val * val;
    }
  }
}

// Construct a const row table for quick access.
template <int T> struct RowTable {
  constexpr RowTable() : values() {
    for (auto i = 1; i < N; i++) {
      values[pow2(i)] = i;
    }
  }
  size_t values[pow2(T)];
};

void print_transversal(uint128_t transversal) {
  constexpr auto table = RowTable<N>();
  for (size_t row = 0; row < N; row++) {
    size_t index = (transversal >> row * N) & ((1 << N) - 1);
    std::cout << table.values[index];
  }
}

int main(int argc, char const *argv[]) {

  CLOptions opts(argv[0]);
  opts.addOption("output,o", "Specify ouput file, otherwise use stdout.",
                 TYPE::STRING);
  opts.addOption("input,i", "Specify input file, otherwise use stdin.",
                 TYPE::STRING);
  opts.addOption("template,T", "Specify the template number", TYPE::INTEGER);
  opts.addOption("ssym,s",
                 "Specify the index of the s square in the template file. This "
                 "defaults to 0.",
                 TYPE::INTEGER);
  opts.addOption("tsym,t",
                 "Specify the index of the t square in the template file. This "
                 "defaults to 1.",
                 TYPE::INTEGER);
  opts.addOption("count,c",
                 "Runs the solver in quite mode. This mode does no produce "
                 "squares but counts the number of squares possible.");
  opts.addOption("pos,p", "Run a hot started template at block position p",
                 TYPE::INTEGER);
  opts.addOption("mod,m", "Specify a modulus", TYPE::INTEGER);
  opts.addOption("class,k", "Specify the class of the given modulus",
                 TYPE::INTEGER);
  opts.addOption("focal,f",
                 "Set a focal point, this focal point will decide for what "
                 "values of p we should use first square solving.",
                 TYPE::INTEGER);
  // Option alllows the use of my templates to match Ian's orbits
  opts.addOption("ian,l",
                 "The file that contains Ian's selection of orbit leaders.",
                 TYPE::STRING);

  CLResults results = opts.parse(argc, argv);
  if (!results.is_valid()) {
    cerr << "Incorrect usage.\n";
    cerr << opts.usage() << "\n";
    return 1;
  }

  if (results["h"]) {
    cout << opts.usage() << "\n";
    return 0;
  }

  ofstream outfile;
  if (results["o"]) {
    outfile.open(any_cast<string>(results["o"].value()));
    if (!outfile.is_open()) {
      cerr << "Failed to open output file "
           << any_cast<string>(results["o"].value()) << ".\n";
      return 2;
    }
  }

  ifstream infile;
  if (results["i"]) {
    infile.open(any_cast<string>(results["i"].value()));
    if (!infile.is_open()) {
      cerr << "Failed to open input file "
           << any_cast<string>(results["i"].value()) << ".\n";
      return 3;
    }
  }

  istream &input = (infile.is_open()) ? infile : cin;
  ostream &output = (outfile.is_open()) ? outfile : cout;

  TemplateOptions tOptions;
  TemplateReader tr(input, tOptions);

  FileOptions fOptions;
  fOptions.squares = 2; // Although there are 3 squares, the writer only uses 2.
  fOptions.bSizes = tOptions.bSizes;
  fOptions.gSizes = tOptions.gSizes;
  fOptions.rIdx = tOptions.rIdx;
  fOptions.cIdx = tOptions.cIdx;
  fOptions.enc = encoding::ALPHA;

  fOptions.ms1 = 0;
  fOptions.ms2 = 1;
  if (results["s"]) {
    try {
      fOptions.ms1 = any_cast<int>(results["s"].value());
    } catch (exception &e) {
      cerr << "Failed to set index of S symbol square.\n";
      return 4;
    }
  }

  if (results["t"]) {
    try {
      fOptions.ms2 = any_cast<int>(results["t"].value());
    } catch (exception &e) {
      cerr << "Failed to set index of T symbol square.\n";
      return 5;
    }
  }

  size_t skip = 0;
  if (results["T"]) {
    try {
      skip = any_cast<int>(results["T"].value());
    } catch (exception &e) {
      cerr << "Failed to get the specified skip value.\n";
      return 6;
    }
  }

  Template tmpl(tOptions.gSizes.size() - 2);
  tr.skip(skip);
  string template_id = tr.nextID();
  tr.read<true>(tmpl);

  if (fOptions.ms1 == fOptions.ms2) {
    cerr << "ssymb and tsymb indices cannot match...\n" << opts.usage() << "\n";
    return 9;
  }

  if (fOptions.ms1 != 0) {
    tmpl.swapSym(fOptions.ms1, 0);
  }

  if (fOptions.ms2 != 1) {
    if (fOptions.ms2 == 0) {
      tmpl.swapSym(1, fOptions.ms1);
    } else {
      tmpl.swapSym(fOptions.ms2, 1);
    }
  }

  size_t pos = 0;
  if (results["p"]) {
    try {
      pos = any_cast<int>(results["p"].value());
    } catch (exception &e) {
      cerr << "Position must be supplied for a hot start.\n"
           << opts.usage() << "\n";
      return 10;
    }
  }

  std::vector<std::array<std::vector<uint128_t>, 2>> orbits =
      tmpl.get_hs_diag_as_uint128_t();

  // We here workout which one of Ian's orbits to use.
  std::vector<uint128_t> orb_leaders;
  if (results["l"]) {
    ifstream ian_stream;
    if (results["l"]) {
      ian_stream.open(any_cast<string>(results["l"].value()));
      if (!ian_stream.is_open()) {
        cerr << "Failed to open input file "
             << any_cast<string>(results["l"].value()) << ".\n";
        return 3;
      }
    }

    std::string line;
    size_t ian_pos = 0;
    while (ian_stream >> line) {
      uint128_t diag = Diag(line).low;
      orb_leaders.push_back(diag);
      ian_pos++;
    }

    std::vector<std::array<std::vector<uint128_t>, 2>> s_orbits(orbits.size());
    for (size_t i = 0; i < orb_leaders.size(); i++) {
      uint128_t leader = orb_leaders[i];

      std::array<std::vector<uint128_t>, 2> orbs;
      for (size_t j = 0; j < orbits.size(); j++) {
        if (std::find(orbits[j][0].begin(), orbits[j][0].end(), leader) != orbits[j][0].end()) {
          orbs = orbits[j];
          break;
        }
      }

      s_orbits[i] = orbs;
    }

    orbits = s_orbits;
  }

  uint128_t orbit_leader = orbits[pos][0][0];

  std::array<std::vector<uint128_t>, 2> junk_diags;
  for (size_t i = 0; i < pos; i++) {
    for (size_t j = 0; j < orbits[i][0].size(); j++) {
      junk_diags[0].push_back(orbits[i][0][j]);
    }

    for (size_t j = 0; j < orbits[i][1].size(); j++) {
      junk_diags[1].push_back(orbits[i][1][j]);
    }
  }

  std::sort(junk_diags[0].begin(), junk_diags[0].end());
  std::sort(junk_diags[1].begin(), junk_diags[1].end());

  // Find the third square diagonals
  std::vector<std::vector<Diag>> diags = gen_diagonals(tmpl.toSquares());
  std::vector<std::vector<uint128_t>> diags_r(diags.size());
  std::transform(diags[0].begin(), diags[0].end(), std::back_inserter(diags_r[0]),
                 [](auto &d) { return d.low; });
  std::transform(diags[1].begin(), diags[1].end(), std::back_inserter(diags_r[1]),
                 [](auto &d) { return d.low; });

  std::sort(diags_r[0].begin(), diags_r[0].end());
  std::sort(diags_r[1].begin(), diags_r[1].end());

  std::vector<std::vector<uint128_t>> diagonals = {{}, {}, {}};

  std::set_difference(diags_r[0].begin(), diags_r[0].end(),
                      junk_diags[0].begin(), junk_diags[0].end(),
                      std::back_inserter(diagonals[0]));
  std::set_difference(diags_r[1].begin(), diags_r[1].end(),
                      junk_diags[1].begin(), junk_diags[1].end(),
                      std::back_inserter(diagonals[1]));
  std::transform(diags[2].begin(), diags[2].end(), std::back_inserter(diagonals[2]),
                 [](auto &d) { return d.low; });

  assert(std::find(diagonals[0].begin(), diagonals[0].end(), orbit_leader) !=
         diagonals[0].end());

  std::cout << "Part: " << pos << "\n";

  std::cout << "(sDiagonals, tDiagonals): (" << diagonals[0].size() << ", "
            << diagonals[1].size() << ")\n";

  size_t mod = 1;
  if (results["m"]) {
    try {
      mod = any_cast<int>(results["m"].value());
    } catch (exception &e) {
      cerr << "Modulus must be an integer.\n" << opts.usage() << "\n";
      return 12;
    }
  }

  if (mod < 1) {
    cerr << "Modulus must be greater than 0.\n";
    return 13;
  }

  std::cout << "(Modulus, class): (" << mod;

  size_t m_class = 0;
  if (results["k"]) {
    try {
      m_class = any_cast<int>(results["k"].value());
    } catch (exception &e) {
      cerr << "Class must be an integer.\n" << opts.usage() << "\n";
      return 14;
    }
  }

  if (m_class >= mod) {
    cerr << "The m_class can be at most the modulus.\n";
    return 15;
  }

  std::cout << ", " << m_class << ")\n";

  std::cout << "Orbit leader: ";
  print_transversal(orbit_leader);
  if (results["l"]) {
    std::cout << " [Ian orbit leader]";
  }
  std::cout << "\n";

  std::cout << "Junk orbits from square 1 were:\n";
  for (size_t i = 0; i < junk_diags[0].size(); i++) {
    print_transversal(junk_diags[0][i]);
    std::cout << "\n";
  }
  if (!junk_diags[0].size()) {
    std::cout << "None\n";
  }

  std::cout << "Junk orbits from square 2 were:\n";
  for (size_t i = 0; i < junk_diags[1].size(); i++) {
    print_transversal(junk_diags[1][i]);
    std::cout << "\n";
  }
  if (!junk_diags[1].size()) {
    std::cout << "None\n";
  }

  FileWriter fw(output, fOptions, template_id);
  size_t solutions = 0;
  if (results["f"]) {
    try {
      size_t f_val = any_cast<int>(results["f"].value());
      if (pos > f_val) {
        solutions = solve<true, true>(fw, diagonals, !results["c"].has_value(),
                                      m_class, mod, orbit_leader);
      } else {
        solutions = solve<true>(fw, diagonals, !results["c"].has_value(),
                                m_class, mod, orbit_leader);
      }
    } catch (exception &e) {
      cerr << "Modulus must be an integer.\n" << opts.usage() << "\n";
      return 12;
    }
  } else {
    solutions = solve<true>(fw, diagonals, !results["c"].has_value(), m_class,
                            mod, orbit_leader);
  }

  std::cout << solutions;
  std::cout << "\n";

  return 0;
}
