#include "solver/solver.hpp"

bool conflict(__uint128_t a, __uint128_t b) { return (a & b) > 0; }

bool intersect(__uint128_t a, __uint128_t b) {
  __uint128_t x = a & b;
  if (x == 0) {
    return false;
  }
  return (x & (x - 1)) == 0;
}

// Switches on the early exit mode for cell tracking.
#define GOOD_ENOUGH_DETECTION 10

#ifdef GOOD_ENOUGH_DETECTION
#define GOOD_LIMIT 10
#endif

std::array<uint128_t, N * N> masks;

struct CellTracker {
  std::vector<std::array<uint128_t, N * N>> sq_cells;
  std::vector<size_t> windows;

  CellTracker(size_t squares) : sq_cells(squares), windows(squares, N * N) {
    for (size_t j = 0; j < sq_cells.size(); j++) {
      for (size_t i = 0; i < N * N; i++) {
        sq_cells[j] = masks;
      }
    }
  }

  void update(uint128_t transversal, size_t square) {
    size_t pos = 0;
    size_t check = windows[square] - 10;

    while (pos < windows[square]) {
      if (conflict(transversal, sq_cells[square][pos])) {
        // uint128_t tmp = sq_cells[square][pos];
        // sq_cells[square][pos] = sq_cells[square][windows[square] - 1];
        // sq_cells[square][windows[square] - 1] = tmp;
        std::swap(sq_cells[square][pos], sq_cells[square][windows[square] - 1]);
        windows[square]--;
        if (windows[square] == check) {
          return;
        }
      } else {
        pos++;
      }
    }
  }

  void revert(size_t square) { windows[square] += N; }

  std::vector<uint128_t> get_cells(size_t square) const {
    std::vector<uint128_t> cell_arr;
    cell_arr.reserve(windows[square]);

    for (size_t idx = 0; idx < windows[square]; idx++) {
      cell_arr.push_back(sq_cells[square][idx]);
    }

    return cell_arr;
  }
};

// Solver options must be defined with for access to multiple
struct SolverOptions {
  CellTracker c_tracker;
  size_t sid = 0;

  SolverOptions(size_t squares) : c_tracker(squares) {}

  bool is_complete(const std::vector<std::vector<uint128_t>> &partial) {
    for (size_t square = 0; square < partial.size(); square++) {
      if (partial[square].size() < N) {
        return false;
      }
    }

    return true;
  }

  std::vector<uint128_t> cellector(const std::vector<uint128_t> &diagonals,
                                   size_t square) {
    std::vector<uint128_t> active_cells = c_tracker.get_cells(square);
    std::vector<uint128_t> min_diags = diagonals;

#ifdef GOOD_ENOUGH_DETECTION
    size_t good_count = 0;
    size_t is_good = false;
#endif

    for (size_t a_idx = 0; a_idx < active_cells.size(); a_idx++) {
      std::vector<uint128_t> diags;
      diags.reserve(min_diags.size() + 1);

      for (size_t d_idx = 0; d_idx < diagonals.size(); d_idx++) {
        if (conflict(active_cells[a_idx], diagonals[d_idx])) {
          diags.push_back(diagonals[d_idx]);
        }

        if (diags.size() >= min_diags.size()) {
          break;
        }
      }

      if (diags.size() == 0) {
        return diags;
      }

      if (diags.size() < min_diags.size()) {
        min_diags = std::move(diags);
#ifdef GOOD_ENOUGH_DETECTION
        if (min_diags.size() < GOOD_ENOUGH_DETECTION) {
          is_good = true;
        }
      }

      if (is_good) {
        good_count++;
        if (good_count == GOOD_LIMIT) {
          return min_diags;
        }
#endif
      }
    }
    return min_diags;
  }

  size_t squarelector(const std::vector<std::vector<uint128_t>> &diagonals,
                      const std::vector<std::vector<uint128_t>> &partial) {

    std::vector<size_t> indices(partial.size(), 0);
    std::iota(indices.begin(), indices.end(), 0);
    return *std::min_element(indices.begin(), indices.end(),
                             [&partial](const auto &a, const auto &b) -> bool {
                               return partial[a].size() < partial[b].size();
                             });
  }

  size_t next_id() { return sid++; }
};

std::vector<Square> squareify(std::vector<std::vector<uint128_t>> &partial) {
  std::vector<Square> squares(partial.size());
  for (size_t square = 0; square < partial.size(); square++) {
    for (size_t cell = 0; cell < N * N; cell++) {
      for (size_t diag = 0; diag < partial[square].size(); diag++) {
        if (conflict(masks[cell], partial[square][diag])) {
          squares[square][cell] = diag + 1;
          break;
        }
      }
    }
  }
  return squares;
}

template <bool WriteSquares>
size_t
solve(FileWriter &fw, const std::vector<std::vector<uint128_t>> &diagonals,
      std::vector<std::vector<uint128_t>> &partial, SolverOptions &s_options) {

  // If the partial is complete then we squareify it for printing.
  if (s_options.is_complete(partial)) {
    if constexpr (WriteSquares) {
      std::vector<Square> squares = squareify(partial);

      fw.write(squares, std::to_string(s_options.next_id()));
    }

    return 1;
  }

  size_t square = s_options.squarelector(diagonals, partial);

  std::vector<uint128_t> cell_diags =
      s_options.cellector(diagonals[square], square);

  size_t accumulator = 0;
  for (size_t d_idx = 0; d_idx < cell_diags.size(); d_idx++) {
    // Add the chosen diag to the partial.
    partial[square].push_back(cell_diags[d_idx]);
    s_options.c_tracker.update(cell_diags[d_idx], square);

    // We build a new set of diagonals.
    std::vector<std::vector<uint128_t>> new_diags(partial.size());
    for (size_t sq = 0; sq < partial.size(); sq++) {
      new_diags.reserve(diagonals[sq].size());

      // If we are looking at the relevant square then we want to eliminate any
      // conflicts. Conversely if we are looking at an off cycle square then we
      // want to keep everything that "intersects" but only exactly once.
      if (square == sq) {
        for (auto &diag : diagonals[square]) {
          if (!conflict(cell_diags[d_idx], diag)) {
            new_diags[square].push_back(diag);
          }
        }
      } else {
        for (auto &diag : diagonals[sq]) {
          if (intersect(cell_diags[d_idx], diag)) {
            new_diags[sq].push_back(diag);
          }
        }
      }
    }

    accumulator += solve<WriteSquares>(fw, new_diags, partial, s_options);

    partial[square].pop_back();
    s_options.c_tracker.revert(square);

    if (accumulator == 100) {
      return accumulator;
    }
  }

  return accumulator;
}
